""" cli

License:
    BSD, see LICENSE.md
"""
import os, sys
from typing import Any, Dict, List, Optional, Tuple, Union
from pathlib import Path
from pprint import pprint
import click
from cocina.constants import cocina_NOT_FOUND, cocina_env_key, cocina_CLI_DEFAULT_HEADER
from cocina.utils import safe_copy_yaml, safe_join
from cocina.config_handler import ConfigArgs, CocinaConfig
from cocina.printer import Printer


#
# CONSTANTS
#
_MISSING_PRINTER_PART: str = "argument 'printer'"
_MISSING_RUN_PART: str = "attribute 'run'"
_MISSING_MAIN_PART: str = "attribute 'main'"


# -------------------------------------------------------------------
# CLI INTERFACE
# -------------------------------------------------------------------
@click.group
@click.pass_context
def cli(ctx):
    ctx.obj = {}


#
# INIT
#
@cli.command(name='init', help='initialize project with .cocina file')
@click.option('--log_dir', '-l',
    type=str,
    required=False,
    help='Log Directory')
@click.option('--package', '-p',
    type=str,
    required=False,
    help='Main Package Name')
@click.option('--force', '-f',
    type=bool,
    required=False,
    is_flag=True)
@click.pass_context
def init(
        ctx,
        log_dir: Optional[str] = cocina_NOT_FOUND,
        package: Optional[str] = cocina_NOT_FOUND,
        force: bool = False):
    src_cocina_path = Path(__file__).parent.parent / 'dot_cocina'
    dest_cocina_path = f'{Path.cwd()}/.cocina'
    safe_copy_yaml(
        src_cocina_path,
        dest_cocina_path,
        log_dir=log_dir,
        constants_package_name=package,
        force=force)
    print(f'cocina: project initialized ({dest_cocina_path})')
    cocina = CocinaConfig.init_for_project()
    pprint(cocina)


#
# JOBS
#
@cli.command(name='job', help='job help text')
@click.argument('jobs', type=str, nargs=-1)
@click.option('--env', '-e', type=str, required=False, help='Environment to run job in')
@click.option('--verbose', '-v',
    type=bool,
    required=False,
    is_flag=True,
    help='Enable verbose output')
@click.option('--dry_run',
    type=bool,
    required=False,
    is_flag=True)
@click.pass_context
def job(
        ctx,
        jobs: str,
        env: Optional[str] = None,
        verbose: bool = True,
        dry_run: bool = False):
    # 1. processs args
    jobs, user_config = _process_jobs_and_user_config(jobs, dry_run)

    # 2. cocina-setup
    cocina, printer = _cocina_printer(*jobs)

    # 3. set environment (if provided)
    if env:
        printer.message(f"Setting environment: {env}")
        os.environ[cocina_env_key] = env

    # 4. run jobs
    for job_name in jobs:
        execute_job(job_name, user_config=user_config, printer=printer)

    # 5. complete
    printer.stop(f"Jobs ({safe_join(jobs)}) completed successfully!")


#
# HELPERS
#
def execute_job(job_name: str, user_config: Optional[dict] = None, printer: Optional[Printer] = None) -> None:
    if printer is None:
        cocina, printer = _cocina_printer(job_name)
    error = False
    try:
        config_args = ConfigArgs(job_name, user_config=user_config)
        printer.message(f'Run Job: {job_name}', vspace=1)
        job_module = config_args.import_job_module()
        try:
            job_module.run(config_args, printer=printer)
        except TypeError as e:
            if _MISSING_PRINTER_PART in str(e):
                job_module.run(config_args)
            else:
                raise e
        except AttributeError as e:
            if _MISSING_RUN_PART in str(e):
                job_module.main()
            else:
                raise e
    except FileNotFoundError as e:
        printer.stop(f"Job configuration not found", error=e)
        sys.exit(1)
    except ImportError as e:
        printer.stop(f"Failed to import job module", error=e)
        sys.exit(1)
    except AttributeError as e:
        if _MISSING_MAIN_PART in str(e):
            printer.stop(f"Job module missing 'run' xor 'main' function", error=e)
            sys.exit(1)
        else:
            printer.stop(f"Likely missing configuration-value", error=e)
            printer.vspace()
            error = e
    if error:
        raise error


#
# INTERNAL
#
def _cocina_printer(
        *name_parts: str,
        cocina: Optional[CocinaConfig] = None) -> Tuple[CocinaConfig, Printer]:
    """Initialize CocinaConfig and Printer instances for CLI operations.

    Args:
        *name_parts: Variable length string arguments to construct log name
        cocina: Existing CocinaConfig instance (creates new if None)
        header: Header string for printer output

    Returns:
        Tuple containing CocinaConfig and Printer instances

    Usage:
        >>> cocina, printer = _cocina_printer('job1', 'task1')
        >>> cocina, printer = _cocina_printer(header='Custom Header')
    """
    if cocina is None:
        cocina = CocinaConfig.init_for_project()
    if name_parts:
        name_parts = safe_join(*name_parts, sep='-')
    printer = Printer(log_dir=cocina.log_dir, log_name_part=name_parts)
    return cocina, printer


def _process_jobs_and_user_config(jobs: List[str], dry_run: bool) -> Tuple[List[str], Dict[str, Any]]:
    """Process job arguments and user configuration from CLI input.

    Args:
        jobs: List of job names and key=value configuration pairs
        dry_run: Whether to enable dry run mode

    Returns:
        Tuple containing list of job names and processed configuration dict

    Usage:
        >>> jobs, config = _process_jobs_and_user_config(['job1', 'key=value'], True)
        >>> jobs, config = _process_jobs_and_user_config(['job1', 'job2'], False)
    """
    _jobs=[]
    _config=dict(DRY_RUN=dry_run)
    for job in jobs:
        if '=' in job:
            k, v = job.split('=')
            _config[k] = v
        else:
            _jobs.append(job)
    _config = _process_user_config(_config)
    return _jobs, _config


def _process_user_config(config: Dict[str, Any]) -> Dict[str, Any]:
    """Process user configuration values with type conversion.

    Args:
        config: Dictionary of configuration key-value pairs

    Returns:
        Processed configuration dictionary with converted values

    Usage:
        >>> result = _process_user_config({'count': '5', 'items': 'a,b,c'})
        >>> result = _process_user_config({'flag': 'true', 'value': '3.14'})
    """
    _config = dict()
    for k, v in config.items():
        if isinstance(v, str):
            if ',' in v:
                v = v.split(',')
                v = [_process_value(x) for x in v if x or (x == 0)]
            else:
                v = _process_value(v)
        _config[k] = v
    return _config


def _process_value(value: str) -> Union[int, float, str]:
    """Convert string value to appropriate numeric type if possible.

    Args:
        value: String value to process and convert

    Returns:
        Converted value as int, float, or original string

    Usage:
        >>> result = _process_value('42')      # Returns: 42 (int)
        >>> result = _process_value('3.14')    # Returns: 3.14 (float)
        >>> result = _process_value('text')    # Returns: 'text' (str)
    """
    try:
        value = float(value)
        if value.is_integer():
            value = int(value)
    except:
        pass
    return value


#
# MAIN
#
cli.add_command(init)
cli.add_command(job)



"""

Cocina Config Handler

This module provides the ConfigHandler class for managing project configuration
using YAML files, constants, and environment variables.

License: BSd 3-clause

"""
#
# IMPORTS
#
import importlib
import os
import sys
import re
from pathlib import Path
from typing import Any, Optional, Union, Self, Sequence
from types import ModuleType
from dataclasses import dataclass, field
from cocina.constants import (
    cocina_CONFIG_FILENAME, YAML_EXT_REGX, cocina_NOT_FOUND,
    ENVIRONMENT_KEYED_IDENTIFIER, cocina_env_key, PY_EXT_REGX
)
from cocina.utils import (
    safe_join, dir_search, read_yaml, replace_dictionary_values,
    keyed_replace_dictionary_values, import_module_from_path
)


#
# HELPERS
#
def cocina_path(
        path: str,
        project_root: str,
        *subfolders: Union[str, int, float],
        ext: Optional[str] = None,
        ext_regex: Optional[str] = None) -> str:
    """Get path of configuration file with flexible path resolution.

    Manages standard paths for a "cocina" project. Here is an example:

        ```python
        path = cocina_path(
            'a.b.c',
            '/path/to/my/project',
            'sub1',
            'sub2',
            ext='.yaml')

        print(path) # ==> /path/to/my/project/sub1/sub2/a/b/c.yaml
        ```

    Note:
        - <path> may or may not include the `project_root`
        - if <path> beings with "/" and does not contain <project_root> the
          the path remains unaltered.
        - otherwise:
            - allows for "dot-paths" (path.to.something) or file paths with "/"
        - <project_root> will be added unless the path begins with a "/"

    Examples with yaml ext and subfolders [config, args]:
        - a.b.job_name (loads config/args/a/b/job_name.yaml)
        - a/b/job_name (loads config/args/a/b/job_name.yaml)
        - a.b.job_name.yml (loads config/args/a/b/job_name.yml)
        - /a/b/job_name (loads /a/b/job_name - absolute path)

    Args:
        path: Path specification (dot or slash separated)
        project_root: Project root directory
        *subfolders: Additional subfolders to include in path
        ext: File extension to add (default: None)
        ext_regex: Regular expression to match and extract extension from path

    Returns:
        Resolved file path
    """
    path = re.sub(f'^{project_root}/', '', path)
    if path[0] == '/':
        return path
    else:
        if ext_regex:
            match = re.search(ext_regex, path)
            if match:
                ext = match.group(0)
                path = re.sub(ext_regex, '', path)
            else:
                ext = ext or ''
        path = re.sub(r'\.', '/', path)
        parts = [project_root] + list(subfolders) + [path]
    return safe_join(*parts, ext=ext)


def get_project_root(project_root: Optional[str] = None) -> str:
    """Get project root directory.

    Searches for .cocina file in parent directories if project_root not provided.

    Args:
        project_root: Optional explicit project root path

    Returns:
        Path to project root directory
    """
    if project_root is None:
        project_root = dir_search(cocina_CONFIG_FILENAME)
    return project_root


#
# DATA CLASSES
#
@dataclass
class CocinaConfig:
    """Dataclass for managing .cocina configuration.

    This dataclass holds all configuration settings loaded from the .cocina file
    that define how Cocina should locate and load configuration files.

    Usage:
        ```python
        cocina = CocinaConfig.init_for_project()
        config_path = cocina.config_folder + '/' + cocina.config_filename
        ```
    """
    config_folder: str
    args_config_folder: str
    config_filename: str
    jobs_folder: str
    constants_module_name: str
    default_env_key: str
    log_dir: Optional[str] = None
    constants_package_name: Optional[str] = None

    @classmethod
    def file_path(cls, project_root: Optional[str] = None) -> Self:
        """gets path for .cocina file

        Args:
            project_root: Optional explicit project root path

        Returns:
            str: project_root/.cocina
        """
        project_root = get_project_root(project_root)
        return f'{project_root}/{cocina_CONFIG_FILENAME}'

    @classmethod
    def init_for_project(cls, project_root: Optional[str] = None) -> Self:
        """Create new CocinaConfig instance for project.

        Loads configuration from .cocina file in project root directory.
        If project_root is None, searches parent directories for .cocina file.

        Args:
            project_root: Optional explicit project root path

        Returns:
            CocinaConfig instance with loaded configuration
        """
        cocina_config = read_yaml(cls.file_path())
        return CocinaConfig(**cocina_config)


@dataclass
class ArgsKwargs:
    """Dataclass with args and kwargs properties.

    Used with ConfigArgs to allow accessing method arguments as
    ca.method_name.args and ca.method_name.kwargs.

    Usage:
        ```python
        args_kwargs = ArgsKwargs.init_from_value({'args': [1, 2], 'kwargs': {'key': 'value'}})
        some_method(*args_kwargs.args, **args_kwargs.kwargs)
        ```
    """
    args: Sequence[Any] = field(default_factory=list)
    kwargs: dict[str, Any] = field(default_factory=dict)

    @staticmethod
    def args_kwargs_from_value(value: Any) -> tuple[list, dict]:
        """Extract args and kwargs from a single value.

        Parsing rules:
        - If value is dict:
            - If keys are exclusively 'args' or 'kwargs': extract values from dict
            - Otherwise: args = [] and kwargs = value
        - If value is list/tuple: args = value, kwargs = {}
        - Otherwise: args = [value], kwargs = {}

        Args:
            value: Value to parse into args and kwargs

        Returns:
            Tuple containing (args_list, kwargs_dict)
        """
        if isinstance(value, dict):
            keys_set = set(value.keys())
            if keys_set.issubset(set(['args', 'kwargs'])):
                args = value.get('args', [])
                kwargs = value.get('kwargs', {})
            else:
                args = []
                kwargs = value
        elif isinstance(value, (list,tuple)):
            args = value
            kwargs = {}
        else:
            args = [value]
            kwargs = {}
        return args, kwargs

    @classmethod
    def init_from_value(cls, value: Any) -> Self:
        """Create ArgsKwargs from a value using args_kwargs_from_value.

        Args:
            value: Value to parse into ArgsKwargs instance

        Returns:
            ArgsKwargs instance with parsed args and kwargs
        """
        args, kwargs = cls.args_kwargs_from_value(value)
        return ArgsKwargs(args=args, kwargs=kwargs)


#
# CLASSES
#
class ConfigHandler:
    """Handle project configuration using YAML files, constants, and environment variables.

    ConfigHandler provides a unified interface for managing configuration data across
    your project. It searches for a `.cocina` configuration file in parent directories,
    loads YAML configuration files, and optionally imports project constants.

    Requirements:
        Your project must have a `.cocina` file in the project root directory. This file
        contains cocina settings and is used to locate the project root and 
        configure how configuration files are loaded.

    Special Value Processing:
        ConfigHandler supports special string patterns in configuration values:
        - `<<KEY_NAME>>`: Replaced with the value of KEY_NAME from the configuration
        - `[[COCINA:ENV]]`: Replaced with the current environment name or stripped if not set

        These patterns enable dynamic configuration values and environment-specific settings.

    Usage:
        ```python
        # Basic usage
        ch = ConfigHandler()
        
        # Access configuration values
        db_host = ch.get('database_host', 'localhost')  # With default
        api_key = ch['api_key']                         # Direct access (raises KeyError if not found)
        log_level = ch.log_level                        # Attribute access
        
        # Check if key exists
        if 'optional_setting' in ch:
            value = ch.optional_setting
        
        # Update configuration
        ch.update(new_setting='value')                    # Keyword arguments
        ch.update({'key': 'value'})                       # Dictionary merge
        ch.update('config/extra.yaml')                    # Load from YAML file

        # Job-specific configuration
        ch.add_job_config('/path/to/job.py')               # Load job config
        ch.add_job_config('/path/to/job.py', version='v2') # Versioned job config
        
        # With constants module
        ch = ConfigHandler('/path/to/my/module')
        ```

    Configuration Loading:
        1. Searches parent directories for `.cocina` file (project root)
        2. Reads `.cocina` configuration settings
        3. Loads main config file (default: config/config.yaml)
        4. Optionally loads environment-specific config based on cocina.ENV_NAME
        5. Imports constants module if module_path provided

    Priority Order (highest to lowest):
        1. Constants from imported module
        2. Configuration values from YAML files
        3. Default values provided to get() method

    Args:
        package_locator: Optional string used to help determine the location of a
            "constants.py" file:
            - None: use cocina.constants_package_name
            - String containing "/":  path to a file withing the same package
              as the constants.py file
            - String NOT containing "/": the name of the package containg the
              constants.py file

    Raises:
        ValueError: If configuration attempts to overwrite constants or .cocina not found
        KeyError: If attempting to access non-existent configuration key without default
    """
    def __init__(self, package_locator: Optional[str] = None, constants: Optional[ModuleType] = None) -> None:
        """Initialize ConfigHandler.
        
        Args:
            package_locator: Optional string used to help determine the location of a
                "constants.py" file:
                - None: use cocina.constants_package_name
                - String containing "/":  path to a file withing the same package
                  as the constants.py file
                - String NOT containing "/": the name of the package containg the
                  constants.py file
            constants: Optional ModuleType. if provided <package_locator> ignored,
                and ch.constants = <constants>
        """
        self.project_root = get_project_root()
        self.cocina = CocinaConfig.init_for_project(self.project_root)
        self.constants = constants or self._import_constants(package_locator)
        self.config, self.environment_name = self._config_and_environment()
        self.config = self.process_values(self.config)
        self._check_protected_keys()

    def update(self, *args: Union[str, dict], **kwargs) -> None:
        """Update configuration with YAML files, dictionaries, or keyword arguments.

        Supports multiple update methods: loading from YAML files (via string paths),
        merging dictionary data, or updating with keyword arguments. All updates
        are validated against protected constants.

        Args:
            *args: Variable arguments supporting:
                - str: Path to YAML file to load and merge (absolute or relative to project root)
                - dict: Dictionary of key-value pairs to merge into configuration
            **kwargs: Key-value pairs to update configuration with

        Raises:
            ValueError: If arguments are invalid type or configuration conflicts with constants
            FileNotFoundError: If YAML file path cannot be found
        """
        for arg in args:
            if isinstance(arg, dict):
                self.config.update(arg)
            elif isinstance(arg, str):
                path = cocina_path(
                    arg,
                    self.project_root,
                    ext_regex=YAML_EXT_REGX,
                    ext='.yaml')
                yaml_config = read_yaml(path, safe=True)
                self.config.update(yaml_config)
            else:
                err = (
                    'ch.update arg must be either '
                    'a string (path to config-file), or '
                    'a dict (key-value pairs to update config)')
                raise ValueError(err)
        self.config.update(kwargs)
        self.config = self.process_values(self.config)
        self._check_protected_keys()

    def process_values(self, config: dict) -> dict:
        """Process configuration values by replacing string keys.

        Given a configuration dict, replace all values that are strings whose
        and whose value is in config-handler-instance.

        This method uses the `keyed_replace_dictionary_values` utility to perform
        dynamic string replacement with special patterns:
        - `<<A_KEY_THAT_EXISTS>>`: Replaced with the value of A_KEY_THAT_EXISTS from the configuration
        - `[[COCINA:ENV]]`: Replaced with the current environment name or stripped if not set

        Args:
            config: Configuration dictionary to process

        Returns:
            Processed configuration dictionary with replaced values
        """
        config = replace_dictionary_values(config, self.config)
        replacements = {ENVIRONMENT_KEYED_IDENTIFIER: self.environment_name}
        config = keyed_replace_dictionary_values(config, **replacements)
        return config

    def get(self, key: str, default: Any = None) -> Any:
        """Get configuration value with default fallback.

        Usage:
            ```python
            c.get('some_key')
            ```

        Note: If value is all uppercase and is a key in config handler,
        returns that value instead (recursive lookup).

        Args:
            key: Key to access
            default: Default value if key not found

        Returns:
            Configuration value or default
        """
        value = getattr(self.constants, key, cocina_NOT_FOUND)
        if value == cocina_NOT_FOUND:
            value = self.config.get(key, default)
        if isinstance(value, str) and value.isupper() and (value in self):
            value = self[value]
        return value

    def __contains__(self, key) -> bool:
        """Check if key exists in configuration or constants.
        
        Usage:
            ```python
            'key' in c
            ```

        Args:
            key: Key to check

        Returns:
            True if key exists, False otherwise
        """
        if isinstance(key, str):
            contains = key in self.config
            if not contains:
                contains = hasattr(self.constants, key)
            return contains
        return False

    def __getitem__(self, key: str) -> Any:
        """Get configuration value (raises error if key does not exist).
        
        Usage:
            ```python
            c['some_key']
            ```

        Args:
            key: Key to access

        Returns:
            Configuration value

        Raises:
            KeyError: If key not found in configuration or constants
        """
        value = self.get(key, cocina_NOT_FOUND)
        if value == cocina_NOT_FOUND:
            raise KeyError(f'{key} not found in config, or constants')
        else:
            return value

    def __getattr__(self, key: str) -> Any:
        """Get configuration value as attribute.
        
        Usage:
            ```python
            c.some_key
            ```

        Args:
            key: Key to access

        Returns:
            Configuration value
        """
        return self.__getitem__(key)

    def __repr__(self) -> str:
        """Return string representation of ConfigHandler."""
        rep = (
                'ConfigHandler:\n'
                f'- constants: {self.constants}\n'
                f'- config: {self.config}')
        return rep

    #
    # INTERNAL
    #
    def _import_constants(self,
            package_locator: Optional[str] = None) -> Union[ModuleType, None]:
        """Import constants module if it exists.
        
        Args:
            package_locator: Optional string used to help determine the location of a
                "constants.py" file:
                - None: use cocina.constants_package_name
                - String containing "/":  path to a file withing the same package
                  as the constants.py file
                - String NOT containing "/": the name of the package containg the
                  constants.py file

        Returns:
            Imported constants module or None
        """
        if package_locator is None:
            package_name = self.cocina.constants_package_name
        elif ('/' in package_locator):
            locator_path = str(Path(package_locator).resolve())
            package_name = re.sub(f'{self.project_root}/', '', locator_path).split('/', 1)[0]
        else:
            package_name = package_locator
        constants_module = None
        if package_name:
            dot_path = f'{package_name}.{self.cocina.constants_module_name}'
            try:
                constants_module = importlib.import_module(dot_path)
            except ImportError as e:
                pass
        return constants_module

    def _config_and_environment(self) -> tuple[dict, Optional[str]]:
        """Load configuration, adding environment-specific config if it exists.
        
        Returns:
            Tuple containing config dictionary and environment name
        """
        config_path = cocina_path(
            self.cocina.config_filename,
            self.project_root,
            self.cocina.config_folder,
            ext_regex=YAML_EXT_REGX,
            ext='.yaml')
        config = read_yaml(config_path, safe=True)
        default_env = config.pop(self.cocina.default_env_key, None)
        environment_name = os.environ.get(cocina_env_key, default_env)
        if environment_name:
            env_path = cocina_path(
                self.project_root,
                self.cocina.config_folder,
                environment_name,
                ext_regex=YAML_EXT_REGX,
                ext='.yaml')
            config.update(read_yaml(env_path, safe=True))
        return config, environment_name

    def _check_protected_keys(self) -> None:
        """Ensure user's config files do not overwrite constants.py values.
        
        Raises:
            ValueError: If configuration attempts to overwrite constants
        """
        if self.constants:
            protected_keys = dir(self.constants)
            protected_keys = [k for k in protected_keys if str(k)[0] != '_']
            if protected_keys:
                for key in protected_keys:
                    config_keys = self.config.keys()
                    if key in config_keys:
                        raise ValueError('Configuration cannot overwrite constants')


class ConfigArgs:
    """Manage job-specific configuration and method arguments.

    ConfigArgs provides a unified interface for loading job configuration files
    and accessing method arguments in a structured way. It integrates with
    ConfigHandler to provide configuration management for job execution.

    The class performs the following operations:
    1. Sets up ConfigHandler for configuration management
    2. Loads argument configuration from specified path
    3. Processes configuration with environment-specific overrides
    4. Creates ArgsKwargs attributes for each method in the config

    Path Resolution:
        - job_name: loads config/args/job_name.yaml
        - a.b.job_name: loads config/args/a/b/job_name.yaml
        - a/b/job_name: loads config/args/a/b/job_name.yaml
        - /a/b/job_name: loads /a/b/job_name (absolute path)

    Warning:
        - Config names cannot include "." except for YAML extension

    Usage:
        ```python
        ca = ConfigArgs('my_job')
        some_method(*ca.some_method.args, **ca.some_method.kwargs)

        # Import and run the job module
        job_module = ca.import_job_module()
        job_module.run(*ca.run.args, **ca.run.kwargs)
        ```
    """
    def __init__(self,
            config_path: Optional[str] = None,
            user_config: Optional[dict] = None,
            config_handler: Optional[ConfigHandler] = None) -> None:
        """Initialize ConfigArgs.

        Args:
            config_path: Path to job configuration file (relative to project_root/config/args/)
            user_config: Optional dictionary of user configuration overrides
            config_handler: Optional existing ConfigHandler instance to use
        """
        # set/load config_handler
        if config_handler:
            self.config_handler = config_handler
        else:
            self.config_handler = ConfigHandler()
        job, args_config, config = self._args_data(config_path, user_config)
        self.config_handler.update(config)
        self.args_config = self.config_handler.process_values(args_config)
        self.property_names = list(self.args_config.keys())
        self.job_path = cocina_path(
            re.sub(YAML_EXT_REGX, '', job or config_path),
            self.config_handler.project_root,
            self.config_handler.cocina.jobs_folder,
            ext='.py',
            ext_regex=PY_EXT_REGX)
        self._set_arg_kwargs()

    def import_job_module(self) -> Any:
        """Helper to import job module.

        Usage:
            ```python
            ca = ConfigArgs(job)
            job_module = ca.import_job_module()
            ```

        Returns:
            Imported job module object
        """
        return import_module_from_path(self.job_path)

    def get(self, key: str, default: Any = None) -> Any:
        """get properties of config_handler and config args as attributes with default fallback.

        Provides attribute-style access to configuration values. First checks
        if the key exists in the config_handler, then falls back to the
        current object's attributes.

        Args:
            key: Attribute name to access

        Returns:
            Value of the requested attribute

        Usage:
            ```python
            ca = ConfigArgs('job_name')
            value = ca.some_config_key  # Access config value as attribute
            ```
        """
        if key in self.config_handler:
            return getattr(self.config_handler, key, default)
        else:
            return getattr(self, key, default)

    def __getattr__(self, key: str) -> Any:
        """Access properties of config_handler and config args as attributes.

        Provides attribute-style access to configuration values. First checks
        if the key exists in the config_handler, then falls back to the
        current object's attributes.

        Args:
            key: Attribute name to access

        Returns:
            Value of the requested attribute

        Usage:
            ```python
            ca = ConfigArgs('job_name')
            value = ca.some_config_key  # Access config value as attribute
            ```
        """ 
        if key in self.config_handler:
            return getattr(self.config_handler, key)
        else:
            return getattr(self, key)

    def __repr__(self) -> str:
        """Return string representation of ConfigHandler."""
        rep = 'ConfigArgs:\n'
        for n in self.property_names:
            rep += f'- {n}: {getattr(self, n)}\n'
        return rep

    #
    # INTERNAL
    #
    def _args_data(self, config_path: str, user_config: Union[dict, None]) -> tuple[dict, Union[str, None], dict]:
        """
        FIX  ME
        """
        args_config_path = cocina_path(
            config_path,
            self.config_handler.project_root,
            self.config_handler.cocina.config_folder,
            self.config_handler.cocina.args_config_folder,
            ext='.yaml',
            ext_regex=YAML_EXT_REGX)
        try:
            args_config = read_yaml(args_config_path)
            # pop special values
            job = args_config.pop('job', None)
            config = args_config.pop('config', {})
            env = args_config.pop('env', {})
            # update ch with config/env
            if self.config_handler.environment_name:
                env = env.pop(self.config_handler.environment_name, {})
                config.update(env)
            if user_config:
                config.update(user_config)
        except Exception:
            job, args_config, config = None, {}, {}
        return job, args_config, config

    def _set_arg_kwargs(self) -> None:
        """Set ArgsKwargs attributes for each property in args_config.

        Creates dynamic attributes on the instance for each key in args_config,
        converting values to ArgsKwargs instances for structured access to
        method arguments and keyword arguments.

        Usage:
            >>> config_args = ConfigArgs('job1')
            >>> config_args._set_arg_kwargs()  # Sets attributes from args_config
            >>> args = config_args.method_name.args
            >>> kwargs = config_args.method_name.kwargs
        """
        for k, v in self.args_config.items():
            setattr(self, k, ArgsKwargs.init_from_value(v))




"""

Cocina Constants

This module contains constant values used throughout the Cocina package.

License: BSd 3-clause

"""
#
# CONSTANTS
#

# cocina:core
cocina_CONFIG_FILENAME: str = '.cocina'
cocina_NOT_FOUND: str = '__cocina_OBJECT_NOT_FOUND'

# cocina:cli
cocina_CLI_DEFAULT_HEADER: str = 'cocina_job'

# REGEX
PY_EXT_REGX: str = r'\.py$'
YAML_EXT_REGX: str = r'\.(yaml|yml)$'
KEY_STR_REGEX: str = r'[a-zA-Z][a-zA-Z0-9_-]*'

# icons
ICON_START = "🚀"
ICON_FAILED = "❌"
ICON_SUCCESS = "✅"

# config/args keyed-identifiers
# - auto-update config/args with env
ENVIRONMENT_KEYED_IDENTIFIER: str = r'\[\[COCINA:ENV\]\]'

# environment variables
# - env-key to store "env-name" to manage environment-specific configs/args
cocina_env_key = "cocina.ENV_KEY"
# - env-key to store path to current log file
cocina_log_path_key = "cocina.LOG_PATH_KEY"




"""

Cocina Utilities

This module contains utility functions and classes for the Cocina package,
including file system operations, YAML handling, and timing functionality.

License: BSd 3-clause

"""
#
# IMPORTS
#
import sys
import importlib.util
import functools
import re
import inspect
import yaml
import json
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, List, Optional, Union
from types import ModuleType
from cocina.constants import cocina_NOT_FOUND, KEY_STR_REGEX


#
# CONSTANTS
#
MAX_DIR_SEARCH_DEPTH: int = 6
TIME_FORMAT: str = '%H:%M:%S'
DATE_TIME_FORMAT: str = '%Y.%m.%d %H:%M:%S'
TIME_STAMP_FORMAT: str = '%Y%m%d-%H%M%S'


#
# IO
#
def read_yaml(path: str, *key_path: str, safe: bool = False) -> Any:
    """Read and optionally extract part of a YAML file.

    # TODO: READ_YAML check for ext?

    Usage:
        ```python
        data = read_yaml(path)
        data_with_key_path = read_yaml(path, 'a', 'b', 'c')
        data['a']['b']['c'] == data_with_key_path  # ==> True
        ```

    Args:
        path: Path to YAML file
        *key_path: Key path to extract from loaded data
        safe: If True, return empty dict when path not found; otherwise raise error

    Returns:
        Dictionary or extracted data from YAML file

    Raises:
        ValueError: If path does not exist and safe=False
    """
    if Path(path).is_file():
        with open(path, 'rb') as file:
            obj = yaml.safe_load(file)
        for k in key_path:
            obj = obj[k]
        return obj
    elif safe:
        return dict()
    else:
        raise ValueError(f'{path} does not exist')


def safe_copy_yaml(
        src: str,
        dest: str,
        force: bool = False,
        **replacements: Any) -> None:
    """Copy yaml files with replacement while preserving comments

    Args:
        src (str): path to source yaml file
        dest (str): destination path for processed yaml file
        force (bool = False): if <dest>-file exists:
            - if <force> overwrite file
            - otherwise raise error
        **replacements (str): key-value pairs to replace value in
            src-yaml-file. if

    Raises:
        ValueError: if <dest> exists and <force> is false

    """
    file_exists = Path(dest).is_file()
    if file_exists and (not force):
        err = f'destination ({dest}) exists. use `force=True` to overwrite file'
        raise ValueError(err)
    else:
        with open(src, "r") as file:
            processed_str = ''
            for line in file:
                for k, v in replacements.items():
                    if (v != cocina_NOT_FOUND) and re.search(f'^{k}', line):
                        line = f'{k}: {json.dumps(v)}'
                    else:
                        line = line.strip()
                processed_str += line + '\n'
        with open(dest, 'w') as file:
            file.write(processed_str)


def write(dest: str, *lines: str, mode='w', strip=True, **kwargs) -> None:
    """Write multiple lines to a file.

    Args:
        dest: Destination file path to write to
        *lines: Variable number of lines to write to the file
        mode: File open mode (default: 'w' for write)
        strip: Whether to strip whitespace from lines (default: True)
        **kwargs: Additional keyword arguments (currently unused)

    Usage:
        >>> write('output.txt', 'Line 1', 'Line 2', 'Line 3')
        >>> write('append.txt', 'New line', mode='a')
    """
    with open(dest, mode) as file:
        for line in lines:
            line = str(line)
            if strip:
                line = line.strip()
            file.write(f'\n{line}')


def import_module_from_path(path: str) -> ModuleType:
    """Import a module from file path.

    Args:
        path: Path to the Python module file to import

    Returns:
        Imported module object
    """
    spec = importlib.util.spec_from_file_location("module.name", path)
    module = importlib.util.module_from_spec(spec)
    sys.modules["module.name"] = module
    spec.loader.exec_module(module)
    return module


#
# INSPECTION
#
def nb_objects(path) -> int:
    """Count the number of objects (files and directories) in a path.

    Args:
        path: Directory path to count objects in

    Returns:
        Number of files and directories in the given path

    Usage:
        ```python
        count = nb_objects('/path/to/directory')
        print(f'Found {count} objects')
        ```
    """
    return len(list(Path(path).glob('*')))


def dir_search(
        *search_names: str,
        max_depth: int = MAX_DIR_SEARCH_DEPTH,
        default: Optional[str] = None) -> str:
    """Search parent directories for files matching search names.

    Args:
        *search_names: File names to search for
        max_depth: Maximum directory depth to search (default: MAX_DIR_SEARCH_DEPTH)
        default: Default return value if files not found (default: None)

    Returns:
        Path to directory containing matching files

    Raises:
        ValueError: If files not found and no default provided
    """
    cwd = Path.cwd()
    directory = cwd
    for _ in range(max_depth):
        file_names = [str(n.name) for n in directory.iterdir()]
        if bool(set(file_names) & set(search_names)):
            return str(directory)
        else:
            directory = directory.parent
    if default:
        return default
    else:
        raise ValueError(f'{search_names} file(s) not found at depth {max_depth}')


def inspect_tree(max_depth: int = 4, sep: str = '.', as_list: bool = False) -> Union[List[str], str]:
    """Inspect the call stack and return function names up to max_depth.

    Args:
        max_depth: Maximum depth of call stack to inspect (default: 4)
        sep: Separator to join function names when returning string (default: '.')
        as_list: Whether to return as list of names or joined string (default: False)

    Returns:
        List of function names if as_list=True, otherwise joined string with separator

    Usage:
        >>> inspect_tree()  # Returns 'caller1.caller2.caller3'
        >>> inspect_tree(as_list=True)  # Returns ['caller1', 'caller2', 'caller3']
        >>> inspect_tree(sep='/')  # Returns 'caller1/caller2/caller3'
    """
    frames = inspect.stack()[1:max_depth+1]
    names = []
    for frame in frames:
        name = frame.function
        if name[0] == '<':
            break
        else:
            names.append(name)
    if as_list:
        return names
    else:
        return safe_join(*names, sep=sep)


def caller_name(max_depth=4, exclude_str='cocina') -> Union[str, None]:
    """Get the name of the calling module by traversing the call stack.

    Traverses the call stack to find the first module name that doesn't
    contain the exclude string. Useful for dynamic header generation.

    Args:
        max_depth: Maximum depth to search in call stack (default: 4)
        exclude_str: String to exclude from module names (default: 'cocina')

    Returns:
        Module name of the caller, or None if not found

    Usage:
        >>> caller_name()  # Returns module name of calling function
        >>> caller_name(max_depth=2, exclude_str='test')
    """
    for i in range(max_depth):
        name = inspect.getmodule(inspect.stack()[i].frame).__name__
        if (name != 'module.name') and (exclude_str not in name):
            return name


#
# CORE
#
def replace_dictionary_values(value: Any, update_dict: dict) -> Any:
    """Replace any values in value-dictionary that are keys in update_dict.

    Recursively traverses dictionaries and lists, replacing string values
    that match keys in update_dict with corresponding values.

    Args:
        value: Dictionary, list, or other value to process
        update_dict: Dictionary containing replacement key-value pairs

    Returns:
        Processed value with replacements applied
    """
    if isinstance(value, dict):
        return {k: replace_dictionary_values(v, update_dict) for k, v in value.items()}
    elif isinstance(value, list):
        return [replace_dictionary_values(v, update_dict) for v in value]
    elif isinstance(value, str):
        return update_dict.get(value, value)
    else:
        return value


def keyed_replace_dictionary_values(
        value: dict,
        open_marker: str = '<<',
        close_marker: str = '>>',
        accepted_regex: str =  KEY_STR_REGEX,
        clean_path: bool =  True,
        **direct_replacements) -> dict:
    """Replace dictionary values using keyed markers and direct replacements.

    Recursively processes a dictionary to replace values containing marker patterns
    with corresponding values from the same dictionary. Also applies direct string
    replacements and optionally cleans up double slashes in paths.

    Args:
        value: Dictionary to process for value replacement
        open_marker: Opening marker pattern (default: '<<')
        close_marker: Closing marker pattern (default: '>>')
        accepted_regex: Regex pattern for valid keys (default: KEY_STR_REGEX)
        clean_path: Whether to clean double slashes in paths (default: True)
        **direct_replacements: Direct key-value string replacements to apply

    Returns:
        Dictionary with replaced values

    Usage:
        ```python
        config = {
            'host': 'localhost',
            'url': 'http://<<host>>/api'
        }
        result = keyed_replace_dictionary_values(config)
        # result['url'] becomes 'http://localhost/api'
        ```
    """
    regex = f'{open_marker}{accepted_regex}{close_marker}'
    value_str = json.dumps(value)
    markers = re.findall(regex, value_str)
    for m in markers:
        key = m.strip(open_marker).strip(close_marker)
        value_str = re.sub(m, value[key], value_str)
    for k, v in direct_replacements.items():
        value_str = re.sub(k, v or '', value_str)
    if clean_path:
        value_str = clean_path_string(value_str)
    return json.loads(value_str)


def clean_path_string(value: str) -> str:
    """
    Removes consecutive forward-slashes "/" from string,
    while preserving protocol delimiters

    Args:
        value: string to be cleaned

    Returns:
        string stripped of multiple slashes
    """
    value = re.sub(r'(://)/+', r'\1', value)
    return re.sub(r'(?<!:)//+', '/', value)


def safe_join(
        *parts,
        sep: str = '/',
        ext: Optional[str] = None,
        clean_path: bool =  True) -> str:
    """Join together non-null values with optional extension.

    Filters out None/empty values and joins remaining parts with separator.
    Optionally adds file extension.

    Args:
        *parts: Parts to join (None/empty values will be filtered out)
        sep: Separator to use for joining (default: '/')
        ext: Optional file extension to add (default: None)
        clean_path: Whether to clean double slashes in paths (default: True)

    Returns:
        Joined string with optional extension
    """
    parts = [str(v) for v in parts if v]
    result = sep.join(parts)
    if ext:
        ext = re.sub(r'^\.', '', ext)
        result = re.sub(f'\\.{ext}$', '', result)
        result = f'{result}.{ext}'
    if clean_path:
        result = clean_path_string(result)
    return result


#
# DECORATORS
#
def singleton(cls):
    """Class decorator that implements the singleton pattern.

    Ensures only one instance of the decorated class can exist. Subsequent
    calls to the class constructor return the same instance.

    Args:
        cls: The class to make singleton

    Returns:
        Function that returns singleton instance

    Usage:
        >>> @singleton
        >>> class MyClass:
        ...     def __init__(self, value):
        ...         self.value = value
        >>>
        >>> obj1 = MyClass(1)
        >>> obj2 = MyClass(2)
        >>> obj1 is obj2  # True - same instance
        >>> obj1.value    # 1 - original value preserved
    """
    instances = {}
    @functools.wraps(cls)
    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    
    return get_instance


#
# DATES AND TIMES
#
class Timer:
    """Simple timer class for measuring elapsed time.

    Usage:
        ```python
        timer = Timer()
        print('Timer starting at:', timer.start())
        print('Start-time as timestamp:', timer.timestamp())
        ...
        print('Current duration:', timer.state())
        ...
        timer.start_lap()
        ...
        print('Duration since start_lap called:', timer.stop_lap())
        ...
        print('Timer stops at:', timer.stop())
        print('Duration that timer ran:', timer.delta())
        ```

    Properties:
        running (bool): Whether the timer is currently running
        initiated (bool): Whether the timer has been started at least once

    Args:
        fmt: Format string for datetime display (default: DATE_TIME_FORMAT)
        ts_fmt: Format string for timestamp display (default: TIME_STAMP_FORMAT)
    """
    def __init__(self, fmt: str = DATE_TIME_FORMAT, ts_fmt: str = TIME_STAMP_FORMAT) -> None:
        self.fmt = fmt
        self.ts_fmt = ts_fmt
        self.start_datetime = None
        self.end_datetime = None
        self.lap_start = None
        self.lap_duration = None
        self.running = False
        self.initiated = False

    def start(self) -> Optional[str]:
        """Start the timer and return formatted start time."""
        self.running = True
        self.initiated = True
        if not self.start_datetime:
            self.start_datetime = datetime.now()
            return self.start_datetime.strftime(self.fmt)
        return None

    def start_lap(self) -> None:
        """Start a lap timer."""
        self.lap_start = datetime.now()

    def stop_lap(self) -> Optional[timedelta]:
        """Stop lap timer and return lap duration.

        Returns:
            Lap duration as timedelta object or None if no lap was started
        """
        if self.lap_start:
            self.lap_duration = datetime.now() - self.lap_start
            self.lap_start = None
            return self.lap_duration
        return None

    def timestamp(self) -> Optional[str]:
        """Return start time as formatted timestamp."""
        if self.start_datetime:
            return self.start_datetime.strftime(self.ts_fmt)
        return None

    def state(self) -> Optional[str]:
        """Return current elapsed time as string."""
        if self.start_datetime:
            return str(datetime.now() - self.start_datetime)
        return None

    def stop(self) -> str:
        """Stop the timer and return formatted stop time.

        Returns:
            Formatted stop time string
        """
        self.running = False
        if not self.end_datetime:
            self.end_datetime = datetime.now()
        return self.end_datetime.strftime(self.fmt)

    def delta(self) -> Optional[str]:
        """Return total elapsed time as string."""
        if self.start_datetime and self.end_datetime:
            return str(self.end_datetime - self.start_datetime)
        return None

    def now(self, fmt: str = 'time') -> str:
        """Return current time in specified format.

        Args:
            fmt: Format type - 'time'/'t' for datetime format, 'timestamp'/'ts' for timestamp format

        Returns:
            Formatted current time string
        """
        if fmt in ['t', 'time']:
            fmt = self.fmt
        elif fmt in ['ts', 'timestamp']:
            fmt = self.ts_fmt
        return datetime.now().strftime(fmt)


"""

Cocina Printer Module

This module provides the Printer class for handling structured output and logging
with timestamps, dividers, and file output capabilities.

License: BSd 3-clause

"""
#
# IMPORTS
#
import re
from pathlib import Path
from typing import Any, Literal, List, Optional, Tuple, Union
from cocina.utils import Timer, safe_join, write, singleton, caller_name
from cocina.constants import (
    ICON_START, ICON_SUCCESS, ICON_FAILED, cocina_CLI_DEFAULT_HEADER,
    cocina_log_path_key
)


#
# CONSTANTS
#
LOG_FILE_EXT: str = 'log'
ERROR_HEADER: str = 'Error'


#
# PUBLIC
#
@singleton
class Printer(object):
    """Structured output and logging class with timestamps and file output.

    Handles formatted printing with headers, timestamps, dividers, and optional
    file logging. Supports timing operations and structured message formatting.
    """
    def __init__(self,
            log_dir: Optional[str] = None,
            log_name_part: Optional[str] = None,
            log_path: Optional[str] = None,
            timer: Optional[Timer] = None,
            basename: Optional[str] = None,
            div_len: int = 100,
            icons: bool = True,
            silent: bool = False,
            start_message: Optional[str] = 'start',
            start_div: Union[str, Tuple[str, str]] = ('=','-'),
            start_vspace: int = 2,
            start_icon: Optional[str] = ICON_START) -> None:
        """Initialize Printer with configuration options.

        Args:
            basename: string to prefix header in messages
            log_dir: Directory path for log file output (optional)
            log_name_part: Part of the log filename to use
            timer: Timer instance for timestamps (creates new if None)
            div_len: Length of divider lines (default: 100)
            icons: Whether to display icons in messages (default: True)
            silent: Whether to suppress console output (default: False)
            start_message: Optional start message to display (default: 'start')
            start_div: Divider characters as string or tuple (default: ('=','-'))
            start_vspace: Vertical spacing before message (default: 2)
            start_icon: icon for start message (None will not include icon)
        Raises:
            ValueError: If header is not a string or list of strings

        Usage:
            >>> printer = Printer(header='MyApp', log_dir='/logs')
            >>> printer = Printer(['Module', 'SubModule'], silent=True)
        """
        self.log_dir = log_dir
        self.log_name_part = log_name_part
        self.log_path = log_path
        self.timer = timer or Timer()
        self.div_len = div_len
        self.icons = icons
        self.silent = silent
        self.basename = basename
        self.timer.start()
        self._process_log_path()
        if start_message:
            self.message(start_message, div=start_div, vspace=start_vspace, icon=start_icon)
        self._initialized = True

    def stop(self,
            message: str = 'complete',
            div: Union[str, Tuple[str, str]] = ('-','='),
            vspace: int = 1,
            error: Union[bool, str, Exception] = False,
             **kwargs: Any) -> str:
        """Stop the printer session and return timing information.

        Args:
            message: Completion message to display (default: 'complete')
            div: Divider characters as string or tuple (default: ('-','='))
            vspace: Vertical spacing before message (default: 1)
            error: Error indicator - False for none, string for custom message, Exception for error object
            **kwargs: Additional keyword arguments passed to message formatting

        Returns:
            Time when stop was called

        Usage:
            >>> stop_time = printer.stop('Processing complete')
            >>> printer.stop('Done', div='#')
        """
        time_stop = self.timer.stop()
        duration = self.timer.delta()
        kwargs['duration'] = duration
        if self.log_path:
            kwargs['log'] = self.log_path
        self.message(
            message,
            div=div,
            vspace=vspace,
            icon=ICON_SUCCESS,
            error=error,
            **kwargs)
        return time_stop

    def message(self,
            msg: str,
            header: Optional[Union[str, Literal[False]]] = None,
            div: Optional[Union[str, Tuple[str, str]]] = None,
            vspace: Union[bool, int] = False,
            icon: Optional[str] = None,
            error: Union[bool, str, Exception] = False,
            callout: bool = False,
            callout_div: str = '*',
            **kwargs: Any) -> None:
        """Print a formatted message with optional dividers and spacing.

        Args:
            msg: Main message content
            header:
                - if None: prefix message with name of module where .message is being called
                - if False: do not prefix message
                - else: prefix message with <header>
            div: Divider characters as string or tuple (optional)
            vspace: Vertical spacing as boolean or number of lines
            icon: Optional icon string to display with message
            error: Error indicator - False for none, string for custom message, Exception for error object
            callout: if callout, wrap message in lines and 2 vertical spaces
            callout_div: character to create wrapping lines
            **kwargs: Additional key-value pairs to append to message

        Usage:
            >>> printer.message('Status update')
            >>> printer.message('Error', 'processing', div='*', vspace=2)
            >>> printer.message('Info', count=42, status='ok')
        """
        if callout:
            self.vspace(2)
            self.line(callout_div)
        self.vspace(vspace)
        if div:
            if isinstance(div, str):
                div1, div2 = div, div
            else:
                div1, div2 = div
            self.line(div1)
        if error:
            error_msg = f'{ERROR_HEADER}:'
            if msg:
                error_msg = f'{error_msg} {msg}'
            if error:
                error_msg = f'{error_msg} [{error}]'
            icon = ICON_FAILED
            msg = error_msg
        if icon and self.icons:
            msg = f'{icon} {msg}'
        self._print(self._format_msg(msg, header, kwargs))
        if div:
            self.line(div2)
        if callout:
            self.line(callout_div)
            self.vspace(2)

    def callout(self,
            msg: str,
            header: Optional[Union[str, Literal[False]]] = None,
            div: Optional[Union[str, Tuple[str, str]]] = None,
            callout_div: str = '*',
            **kwargs: Any) -> None:
        """Convenience wrapper for displaying "callout" messages.

        Makes it easy to see messages in logs by using lines & spacing.
        Mainly for debuging and development

        Args:
            msg: Main message content
            header:
                - if None: prefix message with name of module where .message is being called
                - if False: do not prefix message
                - else: prefix message with <header>
            div: Divider characters as string or tuple (optional)
            **kwargs: Additional key-value pairs to append to message

        Usage:
            >>> printer.message('Status update')
            >>> printer.message('Error', 'processing', div='*', vspace=2)
            >>> printer.message('Info', count=42, status='ok')
        """
        self.message(
            msg,
            header=header,
            div=div,
            callout=True,
            callout_div=callout_div,
            **kwargs)

    def error(self,
            error: Union[bool, str, Exception],
            msg: Optional[str] = None,
            div: Optional[Union[str, Tuple[str, str]]] = None,
            vspace: Union[bool, int] = False,
            icon: Optional[str] = None,
            **kwargs: Any) -> None:
        """Convenience wrapper for displaying error messages.

        Displays an error message with optional formatting. If no message is provided,
        uses the default error message. This method wraps the main message() method
        with error-specific styling.

        Args:
            error: Error condition (bool, string, or Exception)
            msg: Optional error message text
            div: Optional divider formatting
            vspace: Vertical spacing (bool or int)
            icon: Optional icon for the message
            **kwargs: Additional arguments passed to message()

        Usage:
            ```python
            printer = Printer()
            printer.error(True, "Connection failed")
            printer.error(ConnectionError("Timeout"), div="=")
            ```
        """
        self.message(
            msg=msg,
            error=error,
            div=div,
            vspace=vspace,
            icon=icon,
            **kwargs)

    def vspace(self, vspace: Union[Literal[True], int] = True) -> None:
        """Print vertical spacing (blank lines).

        Args:
            vspace: Number of blank lines to print, or False for none

        Usage:
            >>> printer.vspace(2)    # Print 2 blank lines
            >>> printer.vspace(True) # Print 1 blank line
            >>> printer.vspace(False) # Print no blank lines
        """
        if vspace:
            self._print('\n' * int(vspace))

    def line(self, marker: str = '-', length: Optional[int] = None) -> None:
        """Print a horizontal line using repeated marker characters.

        Args:
            marker: Character to repeat for the line (default: '-')
            length: Length of line, uses div_len if None

        Usage:
            >>> printer.line()          # Print line of dashes
            >>> printer.line('=', 50)   # Print line of equals, 50 chars
            >>> printer.line('*')       # Print line of asterisks
        """
        if marker is not None:
            self._print(marker * (length or self.div_len))

    #
    # INTERNAL
    #
    def _process_log_path(self) -> None:
        """Process and set up the log file path for output.

        Determines log file path from environment variables or configuration,
        creates necessary directories, and sets up file logging.

        Usage:
            Internal method called during initialization to configure logging.
        """
        _append = False
        if self.log_path:
            _p =  Path(self.log_path)
            self.log_name = _p.name
            self.log_name_part = _p.stem.split('.')[-1]
            self.log_dir = str(_p.parent)
        elif self.log_dir:
            self.log_name = safe_join(self.timer.timestamp(), self.log_name_part, ext=LOG_FILE_EXT, sep='.')
            self.log_path = safe_join(self.log_dir, self.log_name)
        else:
            self.log_name = None
            self.log_path = None
            self.log_name_part = None
        if self.log_path:
            Path(self.log_path).parent.mkdir(parents=True, exist_ok=True)

    def _format_msg(self, message: str, header: Optional[Union[str, Literal[False]]], key_values: Optional[dict] = None) -> str:
        """Format message with header, timestamp, and key-value pairs.

        Args:
            message: Main message content
            header:
                - if None: prefix message with name of module where .message is being called
                - if False: do not prefix message
                - else: prefix message with <header>
            key_values: Optional dictionary of key-value pairs to append

        Returns:
            Formatted message string with header and timestamp
        """
        if self.timer.initiated:
            timer_part = f'[{self.timer.timestamp()} ({self.timer.state()})] '
        else:
            timer_part = ''
        if header is None:
            header = caller_name()
        header = safe_join(self.basename, header, sep='.')
        if header:
            header = f'{header}: '
        msg = safe_join(timer_part, header, message, sep='')
        if key_values:
            for k,v in key_values.items():
                msg += f'\n\t- {k}: {v}'
        return msg


    def _print(self, message: str) -> None:
        """Print message to console and optionally write to log file.

        Args:
            message: Message string to print and/or log
        """
        if not self.silent:
            print(message)
        if self.log_path:
            write(self.log_path, message, mode='a')
